/******************************************************************************
    SENHA DE 4 DGITOS USANDO TONS DTMF VERSO 2 (SEM INTEGRADO HT8870)

    c/ PIC16F675

    Permite acesso a ambiente por meio de uma senha inserida por meio de tons
    produzidos por um celular na proximidade.
    Trabalha em conjunto com uma fechadura eltrica (porta ou porto).

    Autor: Cludio Lrios

    Baseado nos seguintes projetos e artigos publicados na internet
    e adaptado para linguagem C:

    Tabelas e algoritimos Goertzel modificado: Radu Constantinescu.
    Microchip Note Applications AN257 'DTMF Detection Using PIC18
    Microcontrollers' by Gaurang Kavaiya.
    DTMF Decoding Software with a PIC - by Wim de Vree PE1GRL - pe1grl@amsat.org
    Introduction: DTMF Touch Tone Decoder Using Microchip PIC Microprocessor by
    LargeMouthBass - https://www.instructables.com/member/LargeMouthBass/

    Oscilador:       HS 20MHz xtal
    Compilador Sugerido: CCS C V3.200


    PINAGEM DO PIC 12F675:
    PINO 7 = GP0  sada, led indica pulso dtmf recebido
    PINO 6 = GP1  entrada, COMPARADOR, DTMF audio in
    PINO 5 = GP2  sada, rele para acionar um dispositivo qualquer
    PINO 4 = GP3  (conecte ao vcc com um resistor de 10k)
    PINO 3 = GP4  xtal 20MHZ
    PINO 2 = GP5  xtal 20MHZ

    Senha Padro: tons das teclas telefonicas 1234 (para mudar altere abaixo no
    cdigo e recompile).

    Uso didtico

    Blog Picsource
******************************************************************************/
#define modo_pulso  //comente esta linha para modo retenao

#include <12F675.h>
#use delay(clock=20000000)//frequencia do xtal
#fuses NOWDT,HS, NOCPD, NOPROTECT, NOMCLR, NOPUT, NOBROWNOUT
#rom 0x2100 = {'1','2','3','4'} //senha fixa 1234 (altere conforme desejar)

//bytes

#byte   FSR         = 0x04
#byte   PCLATH      = 0x0A
#byte   STATUS      = 0X03
#byte   TRISIO      = 0x85
#byte   GPIO        = 0x05
#byte   TMR1H       = 0x0F
#byte   TMR1L       = 0x0E
#byte   TMR0        = 0X01
#byte   OPTION_REG  = 0x81
#byte   WPU         = 0x95
#byte   CMCON       = 0x19
#byte   ANSEL       = 0x9F
#byte   ADRESH       = 0x1E
#byte   ADRESL       = 0x9E
#byte   T1CON       = 0X10
#byte   IOC         = 0x96
#byte   INTCON      = 0X0B
#byte   VRCON       = 0X99
#byte   ADCON0      = 0X1F
#byte   VRCON       = 0X99
#byte   FSR         = 0X04
#byte   INDF        = 0X00

//bit
#bit    GIE         = 0xb.7
#bit    PEIE        = 0xb.6
#bit    T0IE        = 0xb.5
#bit    TMR1IE      = 0x8C.0
#bit    TMR1IF      = 0xC.0
#bit    TMR1_ON     = 0x10.0
#bit    T0IF        = 0XB.2
#bit    TMR0_F0     = 0x01.0
#bit    C_OUT       = 0x19.6
#bit    GO_DONE     = 0X1F.1

//pinos
#bit led   = 0x5.0  //pino 7
#bit rele  = 0x5.2  //pino 5


#define N_AMOSTRAGENS       240
#define N_AMOSTRAGENS_X2    113 //1/2 amostragem - 7
#define V_LIMITE_1X_LOW      60  //1/4 amostragem x1
#define V_LIMITE_1X_HIGH     48  //1/5 amostragem x1
#define V_LIMITE_2X_LOW      37  //1/3 amostragem x2 - 3
#define V_LIMITE_2X_HIGH     37  //1/3 amostragem x2 - 3



// variveis globais
char pont_eeprom=0;//endereo do inicio da eeprom
char tom_recebido=0;
char tom_ant=0;
char cnt_pausa=0;
char cnt_amostragem=0;
char mascara;
char tecla_ant=0;
char tecla;

unsigned int1 fvez=0;
/* Acumulador de 32 bytes para guardar valores dos senos e cossenos
das frequncia DTMF -> 697,770,852,941 (baixas) e 1209,1336,1447,1633 (altas)
obtidas de 2 tabelas na flash: tab_low_band e tab_high_band.
Os primeiros 16 bytes contero valores de senos e cossenos para freq x 1.
Os 16 bytes seguintes contero valores de senos e cossenos para freq x 2.
Acesso ao array por nmero 'par' obteremos os valores de senos e 'impares' os
valores de cossenos. Veja abaixo:

sin_cos[0]; -> seno 697 freq x1
sin_cos[1]; -> cosseno 697 freq x1
sin_cos[2]; -> seno 770 freq x1
sin_cos[3]; -> cosseno 770 freq x1
sin_cos[4]; -> seno 852 freq x1
sin_cos[5]; -> cosseno 852 freq x1
sin_cos[6]; -> seno 941 freq x1
sin_cos[7]; -> cosseno 941 freq x1

sin_cos[8]; -> seno 1209 freq x1
sin_cos[9]; -> cosseno 1209 freq x1
sin_cos[10]; -> seno 1336 freq x1
sin_cos[11]; -> cosseno 1336 freq x1
sin_cos[12]; -> seno 1447 freq x1
sin_cos[13]; -> cosseno 1447 freq x1
sin_cos[14]; -> seno 1633 freq x1
sin_cos[15]; -> cosseno 1633 freq x1

sin_cos[16]; -> seno 697 freq x2
sin_cos[17]; -> cosseno 697 freq x2
sin_cos[18]; -> seno 770 freq x2
sin_cos[19]; -> cosseno 770 freq x2
sin_cos[20]; -> seno 852 freq x2
sin_cos[21]; -> cosseno 852 freq x2
sin_cos[22]; -> seno 941 freq x2
sin_cos[23]; -> cosseno 941 freq x2

sin_cos[24]; -> seno 1209 freq x2
sin_cos[25]; -> cosseno 1209 freq x2
sin_cos[26]; -> seno 1336 freq x2
sin_cos[27]; -> cosseno 1336 freq x2
sin_cos[28]; -> seno 1447 freq x2
sin_cos[29]; -> cosseno 1447 freq x2
sin_cos[30]; -> seno 1633 freq x2
sin_cos[31]; -> cosseno 1633 freq x2
*/
unsigned int8 sin_cos[32];//array de senos e cossenos


//tabela de mscaras para as amostragens 697,770,852,941 hz
const char tab_low_band[]={
 0x55,0xff,0xff,0xfa,0xaa,0xaa,0xa0,0x80,0x01,0x05,
 0x15,0x57,0x5f,0x7e,0x7e,0xfa,0xe8,0xe8,0xa0,0xa1,
 0x81,0x87,0x07,0x1f,0x1e,0x5e,0x78,0x78,0x78,0xe1,
 0xe1,0xe1,0x87,0x87,0x86,0x9e,0x1e,0x1c,0x18,0x79,
 0x79,0x71,0x63,0xe7,0xe7,0xc6,0x8e,0x8c,0x9c,0x98,
 0x19,0x39,0x33,0x73,0x63,0x66,0x66,0xe6,0xcc,0xcc,
 0x8d,0x99,0x99,0x9b,0x33,0x32,0x36,0x66,0x64,0x6c,
 0x4c,0xcd,0xc9,0xdb,0x9b,0x93,0x92,0xb2,0x36,0x36,
 0x26,0x6d,0x6d,0x4d,0x4b,0xcb,0xd2,0xd2,0x92,0xb4,
 0xb4,0xb5,0x2d,0x2d,0x2f,0x6b,0x4a,0x4a,0x42,0xd0,
 0xd0,0xd4,0xb5,0xb5,0xbf,0xaf,0x2f,0x2a,0x0a,0x40,
 0x40,0x50,0x55,0xd5,0xd5,0xff,0xbf,0xbe,0xaa,0xaa,
 0x28,0x00,0x01,0x41,0x55,0x57,0x57,0xff,0xfe,0xfa,
 0xf8,0xa8,0xa0,0xa1,0x01,0x07,0x07,0x57,0x5e,0x5e,
 0x7e,0xf8,0xf8,0xe9,0xa1,0xa1,0x83,0x87,0x06,0x06,
 0x1e,0x5c,0x58,0x78,0x79,0xf1,0xe3,0xe3,0xa7,0x96,
 0x86,0x8c,0x1c,0x1c,0x19,0x79,0x79,0x73,0x73,0xe2,
 0xe6,0xe6,0x8c,0x8c,0x8d,0x99,0x19,0x1b,0x33,0x73,
 0x72,0x66,0x64,0xe4,0xcc,0xcd,0x8d,0x8b,0x9b,0x9b,
 0x12,0x32,0x34,0x74,0x64,0x6d,0x6d,0xcd,0xcb,0xcb,
 0x9a,0x92,0x92,0xb0,0x34,0x35,0x35,0x6d,0x6f,0x6f,
 0x4b,0xca,0xca,0xd0,0x90,0x94,0xb5,0xb5,0x3f,0x2f,
 0x2f,0x6a,0x4a,0x4a,0x40,0xc0,0xd1,0xd5,0x95,0xb7,
 0xbf,0xbe,0x2e,0x2a,0x28,0x40,0x40,0x41,0x55,0xd7,
 0xd7,0xff,0xbe,0xbe,0xb8,0xa8,0x28,0x21,0x01,0x41,
 0x47,0x57,0x56,0xde,0xfe
};
//tabela de mscaras para as amostragens 1209,1336,1447,1633 hz
const int8 tab_high_band[]={
 0x55,0xff,0xea,0xa8,0x80,0x05,0x17,0x5f,0x7e,0xf8,
 0xe0,0x81,0x87,0x17,0x1e,0x78,0x78,0xe1,0xe3,0x87,
 0x8e,0x1c,0x38,0x79,0x63,0xe7,0xc6,0x8c,0x9c,0x19,
 0x33,0x73,0x66,0xc4,0xcc,0x99,0x9b,0x33,0x32,0x64,
 0x6c,0xcd,0xcb,0x93,0xb2,0x34,0x24,0x6d,0x4b,0xcb,
 0xd2,0x90,0xb4,0x2d,0x2f,0x4b,0x4a,0xd0,0xd4,0xb5,
 0xbf,0x2f,0x2a,0x40,0x50,0xd5,0xf7,0xbf,0xaa,0x28,
 0x00,0x41,0x57,0xdf,0xfe,0xa8,0xa8,0x01,0x07,0x57,
 0x5e,0xf8,0xf8,0xa1,0xa3,0x07,0x1e,0x5c,0x78,0xf9,
 0xe3,0xa7,0x86,0x0c,0x1c,0x59,0x73,0xe3,0xe6,0x84,
 0x8c,0x19,0x1b,0x73,0x72,0x64,0xcc,0xcd,0x9b,0x9b,
 0x32,0x34,0x64,0x6d,0xcf,0xcb,0x92,0xb0,0x34,0x25,
 0x6f,0x4b,0xca,0xd0,0x94,0xb5,0x3f,0x2f,0x4a,0x40,
 0xd0,0xd5,0xb7,0xbf,0x2a,0x28,0x40,0x41,0xd7,0xff,
 0xbe,0xa8,0x28,0x01,0x47,0x57,0xde,0xfc,0xb8,0xa1,
 0x03,0x06,0x56,0x5c,0xf9,0xf9,0xa3,0xa2,0x06,0x1c,
 0x5d,0x79,0xf3,0xe2,0xa6,0x84,0x0d,0x19,0x5b,0x72,
 0xe2,0xe4,0x8d,0x8d,0x1b,0x1a,0x72,0x74,0xe5,0xed,
 0x8f,0x9a,0x12,0x30,0x75,0x65,0xef,0xca,0x8a,0x90,
 0x11,0x35,0x6f,0x6e,0xca,0xc8,0x91,0x95,0x37,0x3e,
 0x6e,0x68,0xc1,0xd1,0x97,0xbe,0x3e,0x28,0x69,0x41,
 0x47,0xd6,0xfe,0xbc,0xa9,0x21,0x03,0x46,0x56,0xdc,
 0xf9,0xb9,0xa3,0x02,0x06,0x5c,0x5d,0xf9,0xfb,0xa2,
 0xa6,0x04,0x0d,0x5d,0x7b,0xf2,0xe2,0xa4,0x85,0x0d,
 0x1b,0x5a,0x72,0xf4,0xe5,0x8d,0x8f,0x1a,0x12,0x70,
 0x75,0xe5,0xef,0x8a,0x8a
};


/* valores de magnitudes referentes a cada par de frequncia
   sendo :
   bit7=1633  bit6=1447 bit5=1336 bit4=1209
   bit3=941   bit2=852  bit1=770  bit0=697
   Obs; o bit setado indica presena da respectiva frequncia */
const char tab_magnitudes[]={
 //0   1    2    3     4   5    6    7    8    9    *   #     A    B    C    D
  40, 17,  33,  65,   18,  34,  66, 20,   36,  68,  24, 72,  129, 130,  132, 136
};

/* Tabela para retorno de valores de teclas (retorna conforme posies da
tabela de magnitudes */
const char tab_teclas[]={ //                      *      #
 '0','1',  '2', '3', '4', '5','6', '7', '8', '9', 0x2a, 0x23, 'A', 'B','C', 'D'
};

/*==============================================================================
 ROTINA DE INTERRUPO DO TIMER 0 - (executada a cada 102,4 us)
 Obs. O modo de interrupo global no tem salvamento automtico de valores dos
 principais registradores e nem apaga flag de overflow do timer0. Porque usa-la?
 O motivo  que a rotina '#int_timer0' gasta muito tempo salvando registradores
 desnecessrios aos objetivos desta montagem.
 Para funcionar, a rotina de interrupo do timer0 no pode exceder 102,4 us.
 A rotina atual gasta 96,8 us.
 Note que as rotinas usam o FSR diretamente como ponteiro. Tambm foi priorizado
 o uso de rotinas curtas. Cada rotina foi reduzida ao mximo para no exceder
 o tempo mximo de execuo.
 Se fosse usado '#int_timer0' e o modo padro de rotinas C, este tempo passaria
 tranquilamente dos 160 us.
 Cuidado especial deve ser dado a este fato caso venha a migrar para outros
 tipos de microcontroladores.
 Tambm tem que ser usado a frequncia de 20mhz (frequncia mximo deste PIC).
==============================================================================*/
/*==============================================================================
 ROTINA DE INTERRUPO DO TIMER 0 - (executada a cada 102,4 us)
 Obs. O modo de interrupo global no tem salvamento automtico de valores dos
 principais registradores e nem apaga flag de overflow do timer0. Porque usa-la?
 O motivo  que a rotina '#int_timer0' gasta muito tempo salvando registradores
 desnecessrios aos objetivos desta montagem.
 Para funcionar, a rotina de interrupo do timer0 no pode exceder 102,4 us.
 A rotina atual gasta 96,8 us.
 Note que as rotinas usam o FSR diretamente como ponteiro. Tambm foi priorizado
 o uso de rotinas curtas. Cada rotina foi reduzida ao mximo para no exceder
 o tempo mximo de execuo.
 Se fosse usado '#int_timer0' e o modo padro de rotinas C, este tempo passaria
 tranquilamente dos 160 us.
 Cuidado especial deve ser dado a este fato caso venha a migrar para outros
 tipos de microcontroladores.
 Tambm tem que ser usado a frequncia de 20mhz (frequncia mximo deste PIC).
==============================================================================*/
#INT_GLOBAL
void isr_timer0(){
//variveis locais
char cnt_amostragem2x;
char modelo_freq;//guarda modelo da amostragem para uma dada frequencia
char STATUS_TEMP,PCLATH_SAVE,FSR_SAVE,W_TEMP;

   #ASM
   movwf	 W_TEMP
	swapf	 STATUS,W
	clrf	 STATUS		//ENSURE BANK0 SAVE
	movwf	 STATUS_TEMP
	movf	 PCLATH,W	  //SAVE PCLATH
	movwf  PCLATH_SAVE
	movf	 FSR, W
	movwf	 FSR_SAVE
   #ENDASM

  /*testa bit de sada do comparador analgico que contm a informao do sinal
  DTMF (na entrada do pino 6).
  */
  mascara=0;//inicia zerado
  if(C_OUT)mascara--;//Se sada do comparador analgico  '1' ento mascara=0xff
  cnt_amostragem2x=cnt_amostragem<<1;//obtem posio para tabela x2

  /* Obtem valores diferentes de senos e cossenos das frequncias de tom para cada
  amostragem efetuada atravs de 2 tabelas na flash. Estes valores so mascarados
  conforme a situao da entrada de sinal DTMF (pino 12, via comparador analgico).
  */


  FSR= sin_cos; //aponta ponteiro para inicio do acumulador seno/cosseno 32
  modelo_freq=tab_low_band[cnt_amostragem] ^ mascara;//obtem valores das tabela low x1
  do{
    if(modelo_freq & 0x80)INDF--;//bit =1 ento decrementa acumulador apontado
    if(!(modelo_freq & 0x80))INDF++;//bit 0 ento incrementa acumulador apontado
    modelo_freq= modelo_freq <<1; //desloca para esquerda para mudar o bit
    FSR++;//muda o acumulador
    }while(FSR<sin_cos+8);

  modelo_freq=tab_high_band[cnt_amostragem] ^ mascara;//obtem valores das tabela high x1
  do{
    if(modelo_freq & 0x80)INDF--; //se 1 decrementa
    if(!(modelo_freq & 0x80))INDF++;
    modelo_freq= modelo_freq <<1; //desloca para esquerda
    FSR++;
  }while(FSR<sin_cos+16);


  if(cnt_amostragem<N_AMOSTRAGENS_X2){//limita quantidade de amostragens em frequncias 2x
     modelo_freq=tab_low_band[cnt_amostragem2x] ^ mascara;//obtem valores das tabela low x2
     do{
      if(modelo_freq & 0x80)INDF--; //se 1 decrementa
      if(!(modelo_freq & 0x80))INDF++; //se 0 incrementa
      modelo_freq= modelo_freq <<1; //desloca para esquerda
      FSR++;
     }while(FSR<sin_cos+24);
     modelo_freq=tab_high_band[cnt_amostragem2x] ^ mascara;//obtem valores das tabela high x2
     do{
      if(modelo_freq & 0x80)INDF--; //se 1 decrementa
      if(!(modelo_freq & 0x80))INDF++; //se 0 incrementa
      modelo_freq= modelo_freq <<1; //desloca para esquerda
      FSR++;
     }while(FSR<sin_cos+32);
   }


  if(++cnt_amostragem>=N_AMOSTRAGENS){//incrementa e se chegou no valor mximo
    T0IE=0;//Desliga interrupo do timer 0
  }

  T0IF=0;//reseta flag de overflow do timer0

  #asm //restabelece principais registradores com valores anteriores
  	 movf	   FSR_SAVE,W
	 movwf	FSR
	 movf	   PCLATH_SAVE,W
	 movwf	PCLATH		// RESTORE PCLATH
	 swapf	STATUS_TEMP,W
	 movwf	STATUS
	 swapf	W_TEMP,F
	 swapf	W_TEMP,W
  #endasm
}//fim da rotina de interrupo do timer 0


/*------------------------------------------------------------------------------
     ROTINA PARA ZERAR ARRAY COM VALORES DE SENOS E COSSENOS
-------------------------------------------------------------------------------*/
void zera_array(){
  for(FSR=sin_cos;FSR<=sin_cos+32;FSR++){
     INDF=0; //apaga array seno-cosseno
  }
}

/*------------------------------------------------------------------------------
     ROTINA PARA CAPTURAR OS TONS DTMF
-------------------------------------------------------------------------------*/

char get_dtmf(){
char a,b,c,d;
char res_magx1,res_magx2;

   res_magx1=0;//registrador p/ resultados das magnitudes das 8 frequencias x 1
   res_magx2=0;//registrador p/ resultados das magnitudes das 8 frequencias x 2

   for(b=0;b<32;b++){

     d= sin_cos[b];
     if(d & 0x80)d = 0 - d; //obtem o valor absoluto
     b++;
     c= sin_cos[b];
     if(c & 0x80)c = 0 - c; //obtem o valor absoluto
     c= c + d;// soma seno e cosseno absolutos

     if(b<8){//quando b= 1,3,5,7
       res_magx1=res_magx1>>1;
       if(c>=V_LIMITE_1X_LOW)res_magx1=res_magx1 | 0x80;//c>=
     }
     else if(b<16){//quando b= 9,11,13,15
       res_magx1=res_magx1>>1;
       if(c>=V_LIMITE_1X_HIGH)res_magx1=res_magx1 | 0x80;//c>=
     }

     else if(b<24){//quando b= 17,19,21,23
       res_magx2=res_magx2>>1;
       if(c<V_LIMITE_2X_LOW)res_magx2=res_magx2 | 0x80;//c<
     }
     else {//quando b = 25,27,29,31
       res_magx2=res_magx2>>1;
       if(c<V_LIMITE_2X_HIGH)res_magx2=res_magx2 | 0x80;//c<
     }
   }
   
   res_magx1= res_magx1 & res_magx2;// faz um and lgico entre resultados high e low

   for(a=0;a<16;a++){//obtem valor da tecla apertada
     if(res_magx1 == tab_magnitudes[a]){//compara com tabela
       tecla=tab_teclas[a];
       if(!fvez){//primeira recepo de tecla
         fvez=1;
         tecla_ant=tecla;//salva valor para comparao
         return 0;//retorna zero
       }
       else{
         fvez=0;//segunda recepo de tecla
         if(tecla==tecla_ant){//se for a mesma tecla
          return  (tecla);//retorna valor da tecla
         }
       }
     }
   }
   return 0;//tom no encontrado

}


//=============================================================================
//   INICIALIZAO DE PORTS E REGISTRADORES
//=============================================================================
void main (){
  //-----------------------------------------------------
  // PIC 12F675  definies de portas e registradores

  OPTION_REG=0B10000000; //timer 0 presc/2
  WPU=0;//sem pullups
  ANSEL = 0;   // desliga
  TRISIO = 0b111010;  // entradas e sadas
  VRCON  = 0B10100000; //liga vref
  CMCON =  0B00000100;  // liga comparador interno - VREF=0 /NORMAL
  GPIO =   0b00000000;  // desliga sada
  led=1; //pisca led
  delay_ms(500);
  led=0;
  pont_eeprom=0;//zera ponteiro da senha
  tom_ant=0;//zera varivel
  zera_array();//zera acumuladores de senos/cossenos
  TMR0=0;
  INTCON= 0B10100000; //liga interrupo do timer 0 /
//=============================================================================
//   LOOP PRINCIPAL DE REPETIO DO PROGRAMA
//=============================================================================
  for(;;){

   do{//sai da rotina quando receber um tom diferente do anterior ou uma pausa
     while(T0IE);//aguarda finalizao de amostragens
     tom_recebido = get_dtmf();//procura por tom dtmf

     led=0;//mantm led desligado
     if(tom_recebido==0)tom_ant=0;//apaga tom anterior
     if(tom_ant==tom_recebido)tom_recebido=0;//aguarda pausa entre caracteres
     zera_array();//zera array de senos e cossenos
     TMR0=0;//zera timer para dar um tempo
      cnt_amostragem=0;//zera contador de amostragem
     INTCON= 0B10100000; //liga interrupo do timer 0  e inicia nova amostragem
   }while(tom_recebido==0);

     led=1;//indica no led que recebeu tom
     delay_ms(300);//tempo entre tons

 //----------------------------------------------------------------
  //comparao da sequncia de tons recebidos com senha gravada na eeprom
    if(tom_recebido == read_eeprom(pont_eeprom)){
      cnt_pausa=0;//zera contador para impedir de zerar ponteiro da senha
      if(pont_eeprom>=3){
        pont_eeprom=0xff;
        #ifdef modo_pulso
        rele=1;//modo pulso por 2 seg
        delay_ms(2000);
        rele=0;
        #else //modo reteno
        rele=!rele;
        #endif
      }
    }
    else{//tom errado , reseta ponteiro da senha
      pont_eeprom=0xff;
    }
    pont_eeprom++;
    if(pont_eeprom>3)pont_eeprom=0; //se passou de 4 nmeros reseta ponteiro
    tom_ant=tom_recebido;//salva tom recebido para comparao
  }//for(;;)
}//void main()
//-----------------------------------------------------------------------------


